import re
from datetime import datetime

from foledol.django.utils import to_float


class Field:
    def __init__(self, key, label, disabled):
        self.errors = []
        self.key = key
        self.label = label
        self.disabled = disabled
        self.errors = []
        self.value_as_str = ''
        self.external = False

    def set(self, value):
        pass

    def read(self, context, default):
        pass

    def validate(self):
        return True

    def has_error(self):
        return len(self.errors) > 0


class TextField(Field):
    def __init__(self, key, label,
                 placeholder=None,
                 min_length=None,
                 max_length=None,
                 regex=None,
                 blank=True,
                 disabled=False,
                 autocomplete=None,
                 autocomplete_url=None):
        super().__init__(key, label, disabled=disabled)
        self.min_length = min_length
        self.max_length = max_length
        self.placeholder = placeholder
        self.regex = regex
        self.blank = blank
        self.value = None
        self.autocomplete = autocomplete
        self.autocomplete_url = autocomplete_url

    def set(self, value):
        self.value = value if value is not None else None
        self.value_as_str = value if value is not None else ''

    def read(self, context, default):
        self.set(context[self.key] if self.key in context else default)

    def validate(self):
        self.errors = []
        length = len(str(self.value)) if self.value else 0
        if self.value is None and not self.blank:
            self.errors.append("Le champ est obligatoire")
        elif self.min_length and length < self.min_length:
            self.errors.append("La longueur minimale est " + str(self.min_length))
        elif self.max_length and length > self.max_length:
            self.errors.append("La longueur maximale est " + str(self.max_length))
        elif self.regex:
            if not bool(re.search(self.regex, self.value)):
                self.errors.append("Format incorrect")
        return not self.has_error()


class SelectField(Field):
    def __init__(self, key, label, values=None, blank=True, disabled=False, pairs=None):
        super().__init__(key, label, disabled=disabled)
        self.values = values
        self.blank = blank
        self.pairs = pairs
        self.value = None

    def set(self, value):
        self.value = None if value is None else value
        self.value_as_str = str(self.value)

    def read(self, context, default):
        self.set(context[self.key] if self.key in context else default)


class IntegerField(Field):
    def __init__(self, key, label, min_value=None, max_value=None, blank=True, disabled=False):
        super().__init__(key, label, disabled=disabled)
        self.min_value = min_value
        self.max_value = max_value
        self.blank = blank
        self.value = None

    def set(self, value):
        self.value = value if value != '' else None
        self.value_as_str = '' if value is None else str(value)

    def read(self, context, default):
        if self.key in context:
            try:
                self.value = int(context[self.key])
            except ValueError:
                self.value = None
            self.value_as_str = str(context[self.key])
        else:
            self.set(default)

    def validate(self):
        self.errors = []
        if len(self.value_as_str) < 1 and not self.blank:
            self.errors.append("Le champ est obligatoire")
        elif len(self.value_as_str) > 0 and self.value is None:
            self.errors.append("Format incorrect")
        elif self.min_value and self.value < self.min_value:
            self.errors.append("La valeur minimale est " + str(self.min_value))
        elif self.max_value and self.value > self.max_value:
            self.errors.append("La valeur maximale est " + str(self.max_value))
        return not self.has_error()


class FloatField(Field):
    def __init__(self, key, label, min_value=None, max_value=None, blank=True, disabled=False):
        super().__init__(key, label, disabled=disabled)
        self.min_value = min_value
        self.max_value = max_value
        self.blank = blank
        self.value = None

    def set(self, value):
        self.value = value
        self.value_as_str = '' if value is None else str(value).replace('.', ',')

    def read(self, context, default):
        if self.key in context:
            self.value = to_float(context[self.key])
            self.value_as_str = str(context[self.key])
        else:
            self.set(default)

    def validate(self):
        self.errors = []
        if len(self.value_as_str) < 1 and not self.blank:
            self.errors.append("Le champ est obligatoire")
        elif len(self.value_as_str) > 0 and self.value is None:
            self.errors.append("Format incorrect")
        elif self.min_value and self.value < self.min_value:
            self.errors.append("La valeur minimale est " + str(self.min_value))
        elif self.max_value and self.value > self.max_value:
            self.errors.append("La valeur maximale est " + str(self.max_value))
        return not self.has_error()


class DateField(Field):
    def __init__(self, key, label, format="%d/%m/%Y", blank=True, disabled=False):
        super().__init__(key, label, disabled=disabled)
        self.blank = blank
        self.value = None
        self.format = format

    def set(self, value):
        self.value = value
        self.value_as_str = value.strftime(self.format) if value else ''

    def read(self, context, default):
        if self.key in context:
            try:
                self.value = datetime.strptime(context[self.key], self.format)
            except ValueError:
                self.value = None
            self.value_as_str = str(context[self.key])
        else:
            self.set(default)

    def validate(self):
        self.errors = []
        if len(self.value_as_str) < 1 and not self.blank:
            self.errors.append("Le champ est obligatoire")
        elif len(self.value_as_str) > 0 and self.value is None:
            self.errors.append("Format incorrect")
        return not self.has_error()


class BooleanField(Field):
    def __init__(self, key, label, default=False, disabled=False):
        super().__init__(key, label, disabled=disabled)
        self.value = default

    def set(self, value):
        if isinstance(value, str):
            self.value = value == 'on'
        else:
            self.value = value
        self.value_as_str = str(value)

    def read(self, context, default):
        default_value = default if default is not None else False
        self.set(True if self.key in context else default_value)



from django.conf import settings
from django.urls import reverse


def login(test, username, space=settings.DEFAULT_SPACE):
    response = test.client.post(reverse(space + ':user_login'), {'username': username, 'password': 'pwd'}, follow=True)
    test.assertContains(response, 'home')


def set_default(obj, context):
    for key in obj.__dict__.keys():
        if key in context:
            continue
        value = obj.__dict__[key]
        if value is None:
            continue
        context[key] = value


def get_default_path(test, path, space=settings.DEFAULT_SPACE):
    return path if path else reverse(space + ':' + test.prefix + 's')


def get_and_post(test, url, context):
    response = test.client.get(url, context, follow=True)
    test.assertEqual(response.status_code, 200)
    response = test.client.post(url, context, follow=True)
    test.assertEqual(response.status_code, 200)
    return response


def create_object(test, context, ref, obj=None, path=None, space=settings.DEFAULT_SPACE):
    if obj:
        set_default(obj, context)
    context.update({
        'action': 'create',
        'path': ':' + get_default_path(test, path, space)
    })
    response = get_and_post(test, reverse(space + ':' + test.prefix + '_create'), context)
    test.assertContains(response, ref)


def update_object(test, pk, context, ref, path=None, space=settings.DEFAULT_SPACE):
    context.update({
        'action': 'update',
        'path': ':' + get_default_path(test, path, space)
    })
    response = get_and_post(test, reverse(space + ':' + test.prefix + '_update', kwargs={'pk': pk}), context)
    test.assertContains(response, ref)


def delete_object(test, pk, ref, path=None, space=settings.DEFAULT_SPACE):
    context = {
        'action': 'delete',
        'delete': '',
        'path': ':' + get_default_path(test, path, space)
    }
    response = test.client.get(reverse(space + ':' + test.prefix + '_delete', kwargs={'pk': pk}), context, follow=True)
    test.assertEqual(response.status_code, 200)
    response = test.client.post(reverse(space + ':' + test.prefix + '_delete', kwargs={'pk': pk}), context, follow=True)
    test.assertEqual(response.status_code, 200)
    test.assertNotContains(response, ref)


def search_object(test, item1, item2, space=settings.DEFAULT_SPACE):
    response = test.client.post(reverse(space + ':' + test.prefix + 's'), {'search': item1})
    test.assertEqual(response.status_code, 200)
    test.assertContains(response, item1)
    test.assertNotContains(response, item2)
    response = test.client.post(reverse(space + ':' + test.prefix + 's'), {'search': ''})
    test.assertContains(response, item1)
    test.assertContains(response, item2)


def filter_object_all(test, filters, space=settings.DEFAULT_SPACE):
    for filter in filters:
        filter_object(test, filter.name)
    test.client.post(reverse(space + ':' + test.prefix + 's'), {'filter_key': ''})


def filter_object(test, filter_key, space=settings.DEFAULT_SPACE):
    response = test.client.post(reverse(space + ':' + test.prefix + 's'), {'filter_key': filter_key})
    test.assertEqual(response.status_code, 200)


def download(test, action, url, content_type):
    response = test.client.post(url, {
        'action': action,
        'path': ':' + url,
    }, follow=True)
    test.assertEqual(response.status_code, 200)
    test.assertEqual(response['Content-Type'], content_type)
    return response


def print_object(test, action, url):
    return download(test, action, url, 'application/pdf')


def export_object(test, action, url):
    return download(test, action, url, 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')


def export_object_as_txt(test, action, url):
    return download(test, action, url, 'text/plain')


def import_object(test, action, url, path, access='rb'):
    with open(path, access) as file:
        response = test.client.post(url, {
            'action': action,
            'path': ':' + url,
            'upload': file
        }, follow=True)
        test.assertEqual(response.status_code, 200)


def import_object_as_txt(test, action, url, path):
    import_object(test, action, url, path, access='r')


def run_object_module(test, action, url):
    response = test.client.post(url, {
        'action': action,
        'path': ':' + url,
    }, follow=True)
    test.assertEqual(response.status_code, 200)
    return response



import django.db.models


class TableFilter:
    def __init__(self, key, label):
        self.key = key
        self.label = label


class TableButton:
    def __init__(self, label, action):
        self.label = label
        self.action = action


class TableColumn:
    def __init__(self, key, name, type=None, value=None, method=None, link=None, sortable=False):
        self.key = key
        self.name = name
        self.type = type
        self.value = value
        self.method = method
        self.link = link
        self.sortable = sortable


class Table:
    def __init__(self, rows, columns, heading=None, create=None, update=None, search=False, filters=None, buttons=None, placeholder=None):
        self.rows = rows
        if isinstance(rows, django.db.models.QuerySet):
            self.count = rows.count()
        self.columns = columns
        self.heading = heading
        self.create = create
        self.update = update
        self.search = search
        self.filters = filters
        self.buttons = buttons
        self.placeholder = placeholder

    def sort(self, rows):
        pass

    def formatter(self, row):
        return ''

